import numpy as np
import pytest

import psyneulink as pnl
import psyneulink.core.components.functions.stateful.integratorfunctions as Functions
from psyneulink.core.components.functions.function import FunctionError
from psyneulink.core.components.functions.nonstateful.transferfunctions import Angle
from psyneulink.core.globals.parameters import ParameterError

np.random.seed(0)
SIZE=10
test_var = np.random.rand(SIZE)
test_initializer = np.random.rand(SIZE)
test_noise_arr = np.random.rand(SIZE)

RAND0_1 = np.random.random()
RAND2 = np.random.rand()
RAND3 = np.random.rand()

def SimpleIntFun(init, value, iterations, noise, rate, offset, **kwargs):
    assert iterations == 3
    if np.isscalar(noise):
        if "initializer" in kwargs:
            return [4.91845218, 4.78766907, 4.73758993, 5.04920442, 4.09842889,
                    4.2909061,  4.05866892, 5.23154257, 5.23413599, 4.86548903]

        else:
            return [4.12672714, 4.25877415, 4.16954537, 4.12360778, 4.02739283,
                    4.2037768,  4.03845052, 4.39892272, 4.45597924, 3.99547688]
    elif isinstance(noise, pnl.DistributionFunction):
        if "initializer" in kwargs:
            return [6.07047464, 1.45183492, 2.13615798, 3.22296925, 3.29867927,
                    0.9734048, 2.54011924, 3.21213761, 1.54651058, 2.7026355, ]

        else:
            return [5.2787496, 0.92294, 1.56811342, 2.29737262, 3.22764321,
                    0.8862755, 2.51990084, 2.37951776, 0.76835383, 1.83262335]
    else:
        if "initializer" in kwargs:
            return [5.53160614, 4.86244369, 3.79932695, 5.06809088, 2.1305511,
                    3.8879681,  2.16602771, 5.74284825, 4.47697989, 3.78677378]

        else:
            return [4.7398811, 4.33354877, 3.23128239, 4.14249424, 2.05951504,
                    3.8008388, 2.14580932, 4.9102284,  3.69882314, 2.91676163]

def AdaptiveIntFun(init, value, iterations, noise, rate, offset, **kwargs):
    assert iterations == 3

    if np.isscalar(noise):
        if "initializer" in kwargs:
            return [3.44619156, 3.44183529, 3.38970396, 3.49707692, 3.08413924,
                    3.22437653, 3.07231498, 3.66899395, 3.69062231, 3.37774376]
        else:
            return [3.13125441, 3.23144828, 3.16374378, 3.12888752, 3.05588209,
                    3.18971771, 3.06427238, 3.33778941, 3.38108243, 3.03166509]

    elif isinstance(noise, pnl.DistributionFunction):
        if "initializer" in kwargs:
            return [4.18870661, 1.3561085, 1.69287182, 1.94643064, 2.12581409,
                    1.05242466, 2.05628752, 1.90164378, 1.18394637, 1.39578569]

        else:
            return [3.87376946, 1.14572149, 1.46691163, 1.57824123, 2.09755694,
                    1.01776584, 2.04824492, 1.57043925, 0.8744065, 1.04970702]
    else:
        if "initializer" in kwargs:
            return [3.91143701, 3.49857235, 2.67777415, 3.51140748, 1.59096419,
                    2.91863753, 1.63622751, 4.05695955, 3.11611173, 2.55924237]

        else:
            return [3.59649986, 3.28818534, 2.45181396, 3.14321808, 1.56270704,
                    2.88397872, 1.62818492, 3.72575501, 2.80657186, 2.2131637]

def DriftIntFun(init, value, iterations, noise, **kwargs):
    assert iterations == 3

    if np.isscalar(noise):
        if "initializer" not in kwargs:
            return ([0.35782281, 4.03326927, 4.90427264, 0.90944534, 1.45943493,
                     2.31791882, 3.05580281, 1.20089146, 2.8408554 , 1.93964773],
                    [3., 3., 3., 3., 3., 3., 3., 3., 3., 3.])

        else:
            return ([1.14954785, 4.56216419, 5.4723172 , 1.83504198, 1.53047099,
                     2.40504812, 3.07602121, 2.0335113 , 3.61901215, 2.80965988],
                    [3., 3., 3., 3., 3., 3., 3., 3., 3., 3.])

    else:
        if "initializer" not in kwargs:
            return ([0.17810305, 4.06675934, 4.20730295, 0.90582833, 1.60883329,
                     2.27822395, 2.2923697 , 1.10933472, 2.71418965, 1.86808107],
                    [3., 3., 3., 3., 3., 3., 3., 3., 3., 3.])

        else:
            return ([0.96982809, 4.59565426, 4.77534751, 1.83142497, 1.67986935,
                     2.36535325, 2.3125881 , 1.94195457, 3.4923464 , 2.73809322],
                    [3., 3., 3., 3., 3., 3., 3., 3., 3., 3.])

def LeakyFun(init, value, iterations, noise, **kwargs):
    assert iterations == 3

    if np.isscalar(noise):
        if "initializer" not in kwargs:
            return [2.20813608, 2.25674001, 2.22389663, 2.2069879,  2.17157305, 2.23649656, 2.17564317, 2.30832598, 2.32932737, 2.15982541]
        else:
            return [2.93867224, 2.74475902, 2.74803958, 3.06104933, 2.23711905, 2.31689203, 2.19429898, 3.07659637, 3.04734388, 2.96259823]

    elif isinstance(noise, pnl.DistributionFunction):
        if "initializer" not in kwargs:
            return [2.55912037, 1.24455938, 1.43417309, 1.638423, 1.91298882, 1.22700281, 1.71226825, 1.67794471, 1.20395947, 1.48326449]
        else:
            return [3.28965653, 1.73257839, 1.95831604, 2.49248443, 1.97853482, 1.30739828, 1.73092406, 2.4462151, 1.92197598, 2.28603731]

    else:
        if "initializer" not in kwargs:
            return [2.39694798, 2.27976578, 1.9349721, 2.21280371, 1.5655935, 2.11241762, 1.59283164, 2.46577518, 2.09617208, 1.82765063]
        else:
            return [3.12748415, 2.76778478, 2.45911505, 3.06686514, 1.6311395, 2.19281309, 1.61148745, 3.23404557, 2.81418859, 2.63042344]

def AccumulatorFun(init, value, iterations, noise, **kwargs):
    assert iterations == 3

    if np.isscalar(noise):
        if "initializer" not in kwargs:
            # variable is not used in Accumulator
            return [[1.38631136, 1.38631136, 1.38631136, 1.38631136, 1.38631136,
                     1.38631136, 1.38631136, 1.38631136, 1.38631136, 1.38631136]]

        else:
            return [[1.40097107, 1.39610447, 1.39682937, 1.40344986, 1.38762668,
                     1.38792466, 1.38668573, 1.40172829, 1.40071984, 1.40242065]]

    elif isinstance(noise, pnl.DistributionFunction):
        if "initializer" not in kwargs:
            return [[1.46381634, 0.97440038, 0.54931704, 0.28681701, 0.26162584,
                     0.66800459, 1.1010486, 0.02587729, 0.38761176, -0.56452977]]

        else:
            return [[1.47847605, 0.98419348, 0.55983505, 0.30395551, 0.26294116,
                     0.66961789, 1.10142297, 0.04129421, 0.40202024, -0.54842049]]

    else:
        if "initializer" not in kwargs:
            return [[1.65907194, 1.41957474, 0.96892655, 1.39471298, 0.51090402,
                     1.20706503, 0.5443729, 1.61376489, 1.04949166, 0.90644658]]

        else:
            return [[1.67373165, 1.42936784, 0.97944456, 1.41185147, 0.51221934,
                     1.20867833, 0.54474727, 1.62918182, 1.06390014, 0.92255587]]

def DriftOnASphereFun(init, value, iterations, noise, **kwargs):
    assert iterations == 3

    if np.isscalar(noise):
        if "initializer" not in kwargs:
            return [1.10030505e-01,  6.77893188e-06,  4.36876221e-06, -4.83568579e-06,
                    4.55349584e-05,  1.77044532e-04,  1.27797893e-03, -1.92233627e-02,
                    9.74815346e-01, -2.22179738e-01, -6.97708243e-06]

        else:
            return [-1.32269048e-01,  4.35051787e-05,  3.87398441e-05, -3.95620568e-06,
                     1.27324586e-04, -5.01625256e-04, -8.37794371e-04,  1.25048720e-01,
                     7.47570336e-01, -6.52303943e-01, -6.57270465e-05]

    else:
        if "initializer" not in kwargs:
            return [ 0.23690849,  0.00140115,  0.0020072,  -0.00128063,
                    -0.00096267, -0.01620475, -0.02644836,  0.46090672,
                     0.82875571, -0.31584261, -0.00132534]

        else:
            return [-3.72900858e-03, -3.38148799e-04, -6.43154678e-04,  4.36274120e-05,
                     6.67038983e-04, -2.87440868e-03, -2.08163440e-03,  4.41976901e-01,
                     5.31162110e-01, -7.22848147e-01,  4.66808385e-04]


GROUP_PREFIX="IntegratorFunction "

@pytest.mark.function
@pytest.mark.integrator_function
@pytest.mark.parametrize("variable, params", [
    (test_var, {'rate':RAND0_1, 'offset':RAND3}),
    (test_var, {'initializer':test_initializer, 'rate':RAND0_1, 'offset':RAND3}),
    ], ids=["Default", "Initializer"])
@pytest.mark.parametrize("noise", [RAND2, test_noise_arr, pnl.NormalDist],
                         ids=["SNOISE", "VNOISE", "FNOISE"])
@pytest.mark.parametrize("func", [
    (pnl.AdaptiveIntegrator, AdaptiveIntFun),
    (pnl.SimpleIntegrator, SimpleIntFun),
    (pnl.DriftDiffusionIntegrator, DriftIntFun),
    (pnl.LeakyCompetingIntegrator, LeakyFun),
    (pnl.AccumulatorIntegrator, AccumulatorFun),
    pytest.param((pnl.DriftOnASphereIntegrator, DriftOnASphereFun), marks=pytest.mark.llvm_not_implemented),
    ], ids=lambda x: x[0])
@pytest.mark.benchmark
def test_execute(func, func_mode, variable, noise, params, benchmark):
    func_class, func_res = func
    benchmark.group = GROUP_PREFIX + func_class.componentName

    try:
        noise = noise()
    except TypeError as e:
        if "object is not callable" not in str(e):
            raise e from None
    else:
        if issubclass(func_class, (pnl.DriftDiffusionIntegrator, pnl.DriftOnASphereIntegrator)):
            pytest.skip("{} doesn't support functional noise".format(func_class.componentName))

    if issubclass(func_class, pnl.DriftOnASphereIntegrator):
        params = {**params, 'dimension': len(variable) + 1}

    elif issubclass(func_class, pnl.AccumulatorIntegrator):
        params = {**params, 'increment': RAND0_1}
        params.pop('offset', None)

    elif issubclass(func_class, pnl.DriftDiffusionIntegrator):
        # If we are dealing with a DriftDiffusionIntegrator, noise and
        # time_step_size defaults have changed since this test was created.
        # Hard code their old values.
        params = {**params, 'time_step_size': 1.0}
        noise = np.sqrt(noise)

    f = func_class(default_variable=variable, noise=noise, **params)
    ex = pytest.helpers.get_func_execution(f, func_mode)

    ex(variable)
    ex(variable)
    res = benchmark(ex, variable)

    expected = func_res(f.initializer, variable, 3, noise, **params)
    np.testing.assert_allclose(res, expected, rtol=1e-5, atol=1e-8)


def test_integrator_function_no_default_variable_and_params_len_more_than_1():
    I = Functions.AdaptiveIntegrator(rate=[.1, .2, .3])
    I.defaults.variable = np.array([0,0,0])

def test_integrator_function_default_variable_len_1_but_user_specified_and_params_len_more_than_1():
    with pytest.raises(FunctionError) as error_text:
        Functions.AdaptiveIntegrator(default_variable=[1], rate=[.1, .2, .3])
    error_msg_a = 'The length of the array specified for the rate parameter'
    error_msg_b = 'must match the length of the default input'
    assert error_msg_a in str(error_text.value)
    assert error_msg_b in str(error_text.value)

def test_integrator_function_default_variable_and_params_len_more_than_1_error():
    with pytest.raises(FunctionError) as error_text:
        Functions.AdaptiveIntegrator(default_variable=[0,0], rate=[.1, .2, .3])
    error_msg_a = 'The length of the array specified for the rate parameter'
    error_msg_b = 'must match the length of the default input'
    assert error_msg_a in str(error_text.value)
    assert error_msg_b in str(error_text.value)

def test_integrator_function_with_params_of_different_lengths():
    with pytest.raises(FunctionError) as error_text:
        Functions.AdaptiveIntegrator(rate=[.1, .2, .3], offset=[.4,.5])
    error_msg_a = "The parameters with len>1 specified for AdaptiveIntegrator Function"
    error_msg_b = "(['offset', 'rate']) don't all have the same length"
    assert error_msg_a in str(error_text.value)
    assert error_msg_b in str(error_text.value)

def test_integrator_function_with_default_variable_and_params_of_different_lengths():
    with pytest.raises(FunctionError) as error_text:
        Functions.AdaptiveIntegrator(default_variable=[0,0,0], rate=[.1, .2, .3], offset=[.4,.5])
    error_msg_a = "The following parameters with len>1 specified for AdaptiveIntegrator Function"
    error_msg_b = "don't have the same length as its 'default_variable' (3): ['offset']."
    assert error_msg_a in str(error_text.value)
    assert error_msg_b in str(error_text.value)


err_msg_initializer = "'initializer' must be a list or 1d array of length 3 (the value of the 'dimension' parameter minus 1)"
err_msg_angle_func = 'Variable shape incompatibility between (DriftOnASphereIntegrator DriftOnASphereIntegrator'
err_msg_noise = "must be a list or 1d array of length 3 (the value of the 'dimension' parameter minus 1)"

test_vars = [
    ({'initializer': 0.1}, err_msg_initializer, ParameterError),
    ({'initializer': [0.1, 0.1]}, err_msg_initializer, ParameterError),
    ({'initializer': [0.1,0.1,0.1]}, None, None),
    ({'angle_function': Angle}, None, None),
    ({'angle_function': Angle()}, None, None),
    ({'angle_function': Angle([1,1])}, err_msg_angle_func, ParameterError),
    ({'angle_function': Angle([1,1,1])}, None, None),
    ({'noise': .01}, None, None),
    ({'noise': [.01, .5]}, err_msg_noise, FunctionError),
    ({'noise': [.01, .5, .99]}, None, None),
    ({'noise': [.01, .5, .99, .1]}, err_msg_noise, FunctionError)
]

names = [
    "INITIALIZER_SCALAR", "INITIALIZER_2", "INITIALIZER_3",
    "ANGLE_CLASS", "ANGLE_NONE", "ANGLE_2", "ANGLE_3",
    "NOISE_SCALAR", "NOISE_2", "NOISE_3", "NOISE_4"
]


def test_DriftOnASphere_identicalness_against_reference_implementation():
    """Compare against reference implementation in nback-paper model (https://github.com/andrebeu/nback-paper)."""

    # PNL DriftOnASphere
    DoS = Functions.DriftOnASphereIntegrator(dimension=5, initializer=np.array([.2] * (4)), noise=0.0)
    results_dos = []
    for i in range(3):
        results_dos.append(DoS(.1))

    # nback-paper implementation
    def spherical_drift(n_steps=3, dim=5, var=0, mean=.1):
        def convert_spherical_to_angular(dim, ros):
            ct = np.zeros(dim)
            ct[0] = np.cos(ros[0])
            prod = np.prod([np.sin(ros[k]) for k in range(1, dim - 1)])
            n_prod = prod
            for j in range(dim - 2):
                n_prod /= np.sin(ros[j + 1])
                amt = n_prod * np.cos(ros[j + 1])
                ct[j + 1] = amt
            ct[dim - 1] = prod
            return ct
        # initialize the spherical coordinates to ensure each context run begins in a new random location on the unit sphere
        ros = np.array([.2] *(dim - 1))
        slen = n_steps
        ctxt = np.zeros((slen, dim))
        for i in range(slen):
            noise = np.random.normal(mean, var, size=(dim - 1)) # add a separately-drawn Gaussian to each spherical coord
            ros += noise
            ctxt[i] = convert_spherical_to_angular(dim, ros)
        return ctxt
    results_sd = spherical_drift()

    np.testing.assert_allclose(np.array(results_dos), np.array(results_sd))


# FIX: CROSS WITH INITIALIZER SIZE:
@pytest.mark.parametrize("params, error_msg, error_type", test_vars, ids=names)
def test_drift_on_a_sphere_errors(params, error_msg, error_type):
    if error_type:
        with pytest.raises(error_type) as error_text:
            Functions.DriftOnASphereIntegrator(dimension=4, params=params)
        assert error_msg in str(error_text.value)
    else:
        Functions.DriftOnASphereIntegrator(dimension=4, params=params)



import numpy as np
import math
import logging
from numpy import exp, log, sin, sqrt, pi, tanh, cosh, sinh

logger = logging.getLogger(__name__)


def coth(x):
    """
    Hyperbolic cotangent, coth(x) = cosh(x) / sinh(x)
    """
    return cosh(x) / sinh(x)


def __stdWfptLargeTime(t, w, nterms):
    # large time expansion from navarro & fuss
    logger.debug("Large time expansion, %i terms" % nterms)
    piSqOv2 = 4.93480220054

    terms = pi * np.fromiter((k * exp(-k ** 2 * t * piSqOv2) * sin(k * pi * w) for k in range(1, nterms)), np.float64)

    return np.sum(terms)


def __stdWfptSmallTime(t, w, nterms):
    # small time expansion navarro & fuss (todo: gondan et al. improvement)
    logger.debug("Small time expansion, %i terms" % nterms)

    fr = -math.floor((nterms - 1) / 2)
    to = math.ceil((nterms - 1) / 2)

    terms = 1 / sqrt(2 * pi * t ** 3) * np.fromiter(
        ((w + 2 * k) * exp(-((w + 2 * k) ** 2) / (2 * t)) for k in range(fr, to)), np.float64)
    return np.sum(terms)


def wfpt_logp(t, c, x0, t0, a, z, eps=1e-10):
    """
        Log probability of first passage time of double-threshold wiener process
        (aka "pure DDM" of Bogacz et al.). Uses series truncation of Navarro & Fuss 2009
    """

    # boundary sep is 2 * thresh
    boundarySep = 2 * z
    # initial condition is absolute, so to make it relative divide by boundarySep
    relativeInitCond = ((x0 + z) / boundarySep)
    # normalize time
    normT = (t - t0) / (boundarySep ** 2)

    # if t is below NDT, or x0 is outside bounds, probability of any rt is 0
    if normT < 0 or x0 > z or x0 < -z:
        return -np.inf

    if c == 1:  # by default return hit of lower bound, so if resp is correct flip
        a = -a
        relativeInitCond = 1 - relativeInitCond

    largeK = np.int(sqrt((-2 * log(pi * normT * eps)) / (pi ** 2 * normT)))
    smallK = 2 + sqrt(-2 * normT * log(2 * eps * sqrt(2 * pi * normT)))
    if largeK < smallK:
        # make sure eps is small enough for for bound to be valid
        if eps > (1 / (2 * sqrt(2 * pi * normT))):
            smallK = 2
        p = __stdWfptLargeTime(normT, relativeInitCond, math.ceil(largeK))
    else:
        # make sure eps is small enough for for bound to be valid
        if eps > (1 / (pi * sqrt(normT))):
            smallK = math.ceil((1 / (pi * sqrt(t))))
        p = __stdWfptSmallTime(normT, relativeInitCond, math.ceil(smallK))

    # scale from the std case to whatever is our actual
    scaler = (1 / boundarySep ** 2) * exp(-a * boundarySep * relativeInitCond - (a ** 2 * t / 2))

    return scaler * p
    #return log(scaler * p)


def __standardize_bogacz(x0, a, z, s):
    """
        Standardize the way Bogacz et al. 2006 do:
        threshold/drift ratio, signal to noise ratio, x0 to drift ratio
    """
    ztilde = z / a
    atilde = (a / s) ** 2
    x0tilde = x0 / a
    return ztilde, atilde, x0tilde


def __standardize_srivastava(x0, a, z, s):
    """
        Standardize the way Srivastava et al. (submitted) do:
        normalized threshold, normalized starting point
    """
    kz = (a * z) / (s * s)
    kx = (a * x0) / (s * s)
    return kz, kx


def __simulate_wfpt_single(x0, t0, a, z, dt):
    particle = x0
    t = 0;
    while abs(particle) < z:
        particle = particle + np.random.normal(loc=a * dt, scale=sqrt(dt))
        t = t + 1
    return t0 + t * dt if particle > z else -t0 - t * dt


def simulate_wfpt(x0, t0, a, z, dt=0.01):
    """
        Draws from the Wiener first passage time. Slow and imprecise,
        used primarily for testing. Production usage of this function
        is not recommended.
    """
    # promote if we get scalars (probably not the best way to do this)
    x0 = np.atleast_1d(x0)
    t0 = np.atleast_1d(t0)
    a = np.atleast_1d(a)
    z = np.atleast_1d(z)

    return np.fromiter((__simulate_wfpt_single(_x0, _t0, _a, _z, dt) for _x0, _t0, _a, _z in zip(x0, t0, a, z)),
                       np.float64)


def wfpt_rt(x0, t0, a, z, s=1):
    """
    Expected first passage time of a two-boundary wiener process.
    Uses Bogacz et al. 2006 expression for nonzero drift,
    Srivastava et al. expression for zero-drift.
    """
    if abs(a) < 1e-8:  # a close to 0 (avoid float comparison)
        # use expression for limit a->0 from Srivastava et al. 2016
        return t0 + (z ** 2 - x0 ** 2) / (s ** 2)
    # expression from Bogacz et al. 2006 for nonzero drift
    else:
        ztilde, atilde, x0tilde = __standardize_bogacz(x0, a, z, s)
        return ztilde * tanh(ztilde * atilde) + ((2 * ztilde * (1 - exp(-2 * x0tilde * atilde))) / (
                    exp(2 * ztilde * atilde) - exp(-2 * ztilde * atilde)) - x0tilde) + t0


def wfpt_er(x0, t0, a, z, s=1):
    """
    Crossing probability in the -drift direction (aka "Error rate") for wiener process.
    Uses Bogacz et al. 2006 expression for nonzero drift, Srivastava et al.
    expression for zero-drift.
    """
    if abs(a) < 1e-8:  # a close to 0 (avoid float comparison)
        # use expression for limit a->0 from Srivastava et al. 2016
        return (z - x0) / (2 * z)
    # expression from Bogacz et al. 2006 for nonzero drift
    else:
        ztilde, atilde, x0tilde = __standardize_bogacz(x0, a, z, s)
        return 1 / (1 + exp(2 * ztilde * atilde)) - (
                    (1 - exp(-2 * x0 * atilde)) / (exp(2 * ztilde * atilde) - exp(-2 * ztilde * atilde)))


def wfpt_dt_upper(x0, t0, a, z, s=1):
    """
    Expected conditional first passage time, conditioned on crossing threshold
    in the +drift direction (aka "upper" or "correct")
    """

    if abs(a) < 1e-8:  # a close to 0 (avoid float comparison)
        return (4 * x ** 2 - (z + x0) ** 2) / (3 * s ** 2)
    kz, kx = __standardize_srivastava(x0, a, z, s)
    return (s ** 2) / (a ** 2) * (2 * kz * coth(2 * kz) - (kx + kz) * coth(kx + kz))


def wfpt_dt_lower(x0, t0, a, z, s=1):
    """
    Expected conditional first passage time, conditioned on crossing threshold
    in the -drift direction (aka "lower" or "error")
    """
    return wfpt_dt_upper(-x0, t0, a, z, s)


# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# **************************************  CompositionInterfaceMechanism *************************************************

"""

Contents
--------

  * `CompositionInterfaceMechanism_Overview`
  * `CompositionInterfaceMechanism_Creation`
  * `CompositionInterfaceMechanism_Structure`
  * `CompositionInterfaceMechanism_Execution`
  * `CompositionInterfaceMechanism_Class_Reference`


.. _CompositionInterfaceMechanism_Overview:

Overview
--------

CompositionInterfaceMechanisms act as interfaces between a `Composition` and its inputs from and outputs to the
environment, or the Components of another Composition within which it is `nested <Composition_Nested>`.

.. technical_note::

    The CompositionInterfaceMechanism provides both a standard interface through which other Components can interact
    with the environment and/or Compositions, as well as a means of preserving the modularity of Compositions for
    `compilation <Composition_Compilation>`. By providing the standard Components for communication among `Mechanisms
    <Mechanism>` (`InputPorts <InputPort>` and `OutputPorts <OutputPort>`), Mechanisms (and/or other Compositions) that
    are `INPUT <NodeRole.INPUT>` `Nodes <Composition_Nodes>` of a Composition can receive inputs from the environment
    in the same way that any other Node receives inputs, from `afferent Projections <Mechanism_Base.afferents>` (in
    this case, the `input_CIM  <Composition.input_CIM>` of the Composition to which they belong);  and, similarly,
    Components that are `OUTPUT <NodeRole.OUTPUT>` `Nodes <Composition_Nodes>` of a Composition can either report their
    outputs to the Composition or, if they are in a `nested Composition <Composition_Nested>`, send their outputs to
    Nodes in an enclosing Composition just like any others, using `efferent Projections <Mechanism_Base.efferents>`.
    Similarly, for Compilation, they provide a standard interface through which to provide inputs to a Composition and
    for aggregating outputs that, again, maintain a standard interface to other Components (which may not be compiled).

.. _CompositionInterfaceMechanism_Creation:

Creation
--------

The following three CompositionInterfaceMechanisms are created and assigned automatically to a Composition when it is
constructed (and should never be constructed manually):  `input_CIM <Composition.input_CIM>`, `parameter_CIM
<Composition.parameter_CIM>` and `output_CIM <Composition.output_CIM>` (see `Composition_CIMs` for additional details).

.. _CompositionInterfaceMechanism_Structure:

Structure
---------

A CompositionInterfaceMechanisms has a set of `InputPort` / `OutputPort` pairs that its `function
<Mechanism_Base.function>` -- the `Identity` `Function` -- uses to transmit inputs to CompositionInterfaceMechanism
to its outputs.  These are listed in its `port_map  <CompositionInterfaceMechanism.port_map>` attribute, each entry
of which is a key designating the `Port` of the Component with which the CompositionInterfaceMechanism communicates
outside the Composition (i.e., from an `input_CIM <Composition.input_CIM>` receives an `afferent Projection
<Mechanism_Base.afferents>`, a `parameter_CIM <Composition.parameter_CIM>` receives a `modulatory projection
<ModulatoryProjections>`, or an `output_CIM <Composition.output_CIM>` sends an `efferent Projection
<Mechanism_Base.efferents>`), and the value of which is a tuple containing the corresponding (`InputPort`,
`OutputPort`) pair used to transmit the information to or from the CompositionInterfaceMechanism.
CompositionIntefaceMechanisms can be seen graphically using the `show_cim <ShowGraph.show_cim>` option of the
Composition's `show_graph <ShowGraph.show_graph>` method (see figure below).

.. figure:: _static/CIM_figure.svg

   **Examples of Projections to nested Compositions routed through CompositionInterfaceMechanisms.**  *Panel A:*
   Simple example showing a basic configuration.  *Panel B:*  More complex configuration, generated from script below,
   showing Projections automatically created from the Node of an outer Composition (*X*) to two `INPUT
   <NodeRole.INPUT>` `Nodes <Composition_Nodes>` of a `nested Composition <Composition_Nested>`, a `ControlProjection`
   from a `ControlMechanism` in the outer Composition to a Node it modulates in the nested one, and from a `PROBE
   <NodeRole.PROBE>` Node (*B*) in the nested Composition to the `ControlMechanism` that monitors it. ::

    A = ProcessingMechanism(name='A')
    B = ProcessingMechanism(name='B')
    C = ProcessingMechanism(name='C')
    D = ProcessingMechanism(name='D')
    E = ProcessingMechanism(name='E')
    F = ProcessingMechanism(name='F')
    nested_comp = Composition(pathways=[[A,B,C], [D,E,F]], name='NESTED COMPOSITION')
    X = ProcessingMechanism(name='INPUT NODE')
    Y = ProcessingMechanism(name='OUTPUT NODE')
    C = ControlMechanism(name='CONTROL MECHANISM',
                         monitor_for_control=B,
                         control=("slope", E))
    outer_comp = Composition(name='OUTER COMPOSITION', pathways=[X, nested_comp, Y, C])
    outer_comp.show_graph(show_cim=NESTED, show_node_structure=True)

.. _CompositionInterfaceMechanism_Execution:

Execution
---------

A CompositionInterface Mechanism is executed when the Composition to which it belongs is executed, and shown never
be executed manually.

.. _CompositionInterfaceMechanism_Class_Reference:

Class Reference
---------------

"""

import warnings
from collections.abc import Iterable

from beartype import beartype

from psyneulink._typing import Optional, Union

from psyneulink.core.components.functions.nonstateful.transferfunctions import Identity
from psyneulink.core.components.mechanisms.mechanism import Mechanism
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism_Base
from psyneulink.core.components.ports.inputport import InputPort
from psyneulink.core.components.ports.modulatorysignals.controlsignal import ControlSignal
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.globals.context import ContextFlags, handle_external_context
from psyneulink.core.globals.keywords import COMPOSITION_INTERFACE_MECHANISM, INPUT_PORTS, OUTPUT_PORTS, \
    PREFERENCE_SET_NAME
from psyneulink.core.globals.parameters import Parameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet, REPORT_OUTPUT_PREF
from psyneulink.core.globals.preferences.preferenceset import PreferenceEntry, PreferenceLevel

__all__ = ['CompositionInterfaceMechanism']


class CompositionInterfaceMechanism(ProcessingMechanism_Base):
    """
    CompositionInterfaceMechanism(  \
        function=Identity())

    Subclass of `ProcessingMechanism <ProcessingMechanism>` that acts as interface between a Composition and its inputs
    from and outputs to the environment or other Components (if it is a `nested Composition <Composition_Nested>`).

    See `Mechanism <Mechanism_Class_Reference>` for arguments and additional attributes.

    Attributes
    ----------

    function : InterfaceFunction : default Identity
        the function used to transform the variable before assigning it to the Mechanism's OutputPort(s)

    port_map : dict[Port:(InputPort,OutputPort)]
        entries are comprised of keys designating a Component outside the Composition with which it communicates,
        and values tuples that designate the corresponding `InputPort` - `OutputPort` pairs used to transmit that
        information into or out of the Composition (see `CompositionInterfaceMechanism_Structure`, and
        `Composition_CIMs` under Composition for additional details).
    """

    componentType = COMPOSITION_INTERFACE_MECHANISM
    outputPortTypes = [OutputPort, ControlSignal]

    classPreferenceLevel = PreferenceLevel.TYPE
    # These will override those specified in TYPE_DEFAULT_PREFERENCES
    classPreferences = {
        PREFERENCE_SET_NAME: 'CompositionInterfaceMechanismCustomClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE)}

    class Parameters(ProcessingMechanism_Base.Parameters):
        """
            Attributes
            ----------

                function
                    see `function <CompositionInterfaceMechanism.function>`

                    :default value: `Identity`
                    :type: `Function`
        """
        function = Parameter(Identity, stateful=False, loggable=False)

    @check_user_specified
    @beartype
    def __init__(self,
                 default_variable=None,
                 input_shapes=None,
                 input_ports: Optional[Union[Iterable, Mechanism, OutputPort, InputPort]] = None,
                 function=None,
                 composition=None,
                 port_map=None,
                 params=None,
                 name=None,
                 prefs:   Optional[ValidPrefSet] = None):

        self.composition = composition
        self.port_map = port_map
        self.connected_to_composition = False
        self.user_added_ports = {
            INPUT_PORTS: set(),
            OUTPUT_PORTS: set()
        }
        super(CompositionInterfaceMechanism, self).__init__(default_variable=default_variable,
                                                            input_shapes=input_shapes,
                                                            input_ports=input_ports,
                                                            function=function,
                                                            params=params,
                                                            name=name,
                                                            prefs=prefs,
                                                            )

    @handle_external_context()
    def add_ports(self, ports, context=None):
        ports = super(CompositionInterfaceMechanism, self).add_ports(ports, context=context)
        if context.source == ContextFlags.COMMAND_LINE:
            # warnings.warn(
            #     'You are attempting to add custom ports to a CIM, which can result in unpredictable behavior and '
            #     'is therefore recommended against. If suitable, you should instead add ports to the mechanism(s) '
            #     'that project to or are projected to from the CIM.')
            # if ports[INPUT_PORTS]:
            #     self.user_added_ports[INPUT_PORTS].update([port for port in ports[INPUT_PORTS].data])
            # if ports[OUTPUT_PORTS]:
            #     self.user_added_ports[OUTPUT_PORTS].update([port for port in ports[OUTPUT_PORTS].data])
            from psyneulink.core.compositions.composition import CompositionError
            raise CompositionError(f"Adding ports to a {self.__class__.__name__} is not supported at this time; "
                                   f"these are handled automatically when a Composition is created.")
        return ports

    @handle_external_context()
    def remove_ports(self, ports, context=None):
        super(CompositionInterfaceMechanism, self).remove_ports(ports, context)
        input_ports_marked_for_deletion = set()
        for port in self.user_added_ports[INPUT_PORTS]:
            if port not in self.input_ports:
                input_ports_marked_for_deletion.add(port)
        self.user_added_ports[INPUT_PORTS] = self.user_added_ports[INPUT_PORTS] - input_ports_marked_for_deletion
        output_ports_marked_for_deletion = set()
        for port in self.user_added_ports[OUTPUT_PORTS]:
            if port not in self.output_ports:
                output_ports_marked_for_deletion.add(port)
        self.user_added_ports[OUTPUT_PORTS] = self.user_added_ports[OUTPUT_PORTS] - output_ports_marked_for_deletion

    # def _get_source_node_for_input_CIM(self, port, start_comp=None, end_comp=None):
    def _get_source_node_for_input_CIM(self, port, start_comp=None)->tuple or None:
        """Return Port, Node and Composition  for source of projection to input_CIM from (possibly nested) outer comp
        Return None if there is no source Node (i.e., port receives input direcdtly from input to outermost Composition)
        **port** InputPort or OutputPort of the input_CIM from which the local projection projects;
        **comp** Composition at which to begin the search (or continue it when called recursively;
                 assumes the current CompositionInterfaceMechanism's Composition by default
        """
        # Ensure method is being called on an output_CIM
        assert self == self.composition.input_CIM
        #  CIM MAP ENTRIES:  [RECEIVER InputPort,  [input_CIM InputPort,  input_CIM OutputPort]]
        # Get sender to input_port of input_CIM
        comp = start_comp or self.composition
        port_map = port.owner.port_map
        idx = 0 if isinstance(port, InputPort) else 1
        input_port = [port_map[k][0] for k in port_map if port_map[k][idx] is port]
        assert len(input_port)==1, f"PROGRAM ERROR: Expected exactly 1 input_port for {port.name} " \
                                   f"in port_map for {port.owner}; found {len(input_port)}."
        if not len(input_port[0].path_afferents):
            return None
        sender = input_port[0].path_afferents[0].sender
        if not isinstance(sender.owner, CompositionInterfaceMechanism):
            return sender, sender.owner, comp
        return self._get_source_node_for_input_CIM(sender, sender.owner.composition)

    def _get_destination_info_from_input_CIM(self, port, comp=None):
        """Return Port, Node and Composition for "ultimate" destination of projection to **port**.
        **port**: InputPort or OutputPort of the input_CIM to which the projection of interest projects;
                  used to find destination (key) in output_CIM's port_map.
        **comp**: Composition at which to begin the search (or continue it when called recursively);
                 assumes the Composition for the input_CIM to which **port** belongs by default
        """
        # Ensure method is being called on an input_CIM
        assert self == self.composition.input_CIM
        #  CIM MAP ENTRIES:  [RECEIVER PORT,  [input_CIM InputPort,  input_CIM OutputPort]]
        # Get receiver of output_port of input_CIM
        comp = comp or self.composition
        port_map = port.owner.port_map
        idx = 0 if isinstance(port, InputPort) else 1
        output_ports = [port_map[k][1] for k in port_map if port_map[k][idx] is port]
        assert len(output_ports)==1, f"PROGRAM ERROR: Expected exactly 1 output_port for {port.name} " \
                                   f"in port_map for {port.owner}; found {len(output_ports)}."
        output_port = output_ports[0]
        assert len(output_port.efferents)==1, f"PROGRAM ERROR: Port ({output_port.name}) expected to have " \
                                              f"just one efferent; has {len(output_port.efferents)}."
        receiver = output_port.efferents[0].receiver
        if not isinstance(receiver.owner, CompositionInterfaceMechanism):
            return receiver, receiver.owner, comp
        return self._get_destination_info_from_input_CIM(receiver, receiver.owner.composition)

    def _get_modulated_info_from_parameter_CIM(self, port, comp=None):
        """Return Port, Node and Composition for parameter modulated by ControlSignal that projects to parameter_CIM.
        **port**: InputPort or OutputPort of the parameter_CIM to which the ControlSignal projects;
                  used to find destination (key) in parameter_CIM's port_map.
        **comp**: Composition at which to begin the search (or continue it when called recursively);
                 assumes the Composition for the parameter_CIM to which **port** belongs by default.
        """
        # Ensure method is being called on a parameter_CIM
        assert self == self.composition.parameter_CIM
        #  CIM MAP ENTRIES:  [RECEIVER PORT,  [input_CIM InputPort,  input_CIM OutputPort]]
        # Get receiver of output_port of input_CIM
        comp = comp or self.composition
        port_map = port.owner.port_map
        idx = 0 if isinstance(port, InputPort) else 1
        output_port = [port_map[k][1] for k in port_map if port_map[k][idx] is port]
        assert len(output_port)==1, f"PROGRAM ERROR: Expected exactly 1 output_port for {port.name} " \
                                   f"in port_map for {port.owner}; found {len(output_port)}."
        assert len(output_port[0].efferents)==1, f"PROGRAM ERROR: Port ({output_port.name}) expected to have " \
                                                 f"just one efferent; has {len(output_port.efferents)}."
        receiver = output_port[0].efferents[0].receiver
        if not isinstance(receiver.owner, CompositionInterfaceMechanism):
            return receiver, receiver.owner, comp
        return self._get_modulated_info_from_parameter_CIM(receiver, receiver.owner.composition)

    def _get_source_of_modulation_for_parameter_CIM(self, port, comp=None):
        """Return ControlSignal, Node and Composition that projects to a ParameterPort from (possibly nested) outer comp
        **port**: InputPort or OutputPort of the parameter_CIM from which the local ControlSignal projects;
                  used to find source (key) in parameter_CIM's port_map.
        **comp**: Composition at which to begin the search (or continue it when called recursively);
                 assumes the Composition for the parameter_CIM to which **port** belongs by default.
        """
        # Ensure method is being called on a parameter_CIM
        assert self == self.composition.parameter_CIM
        #  CIM MAP ENTRIES:  [RECEIVER ParameterPort : (parameter_CIM InputPort,  parameter_CIM ControlSignal)
        # Get sender to input_port of parameter_CIM
        comp = comp or self.composition
        port_map = port.owner.port_map
        idx = 0 if isinstance(port, InputPort) else 1
        input_port = [port_map[k][0] for k in port_map if port_map[k][idx] is port]
        assert len(input_port)==1, f"PROGRAM ERROR: Expected exactly 1 input_port for {port.name} " \
                                   f"in port_map for {port.owner}; found {len(input_port)}."
        assert len(input_port[0].path_afferents)==1, f"PROGRAM ERROR: Port ({input_port.name}) expected to have " \
                                                     f"just one path_afferent; has {len(input_port.path_afferents)}."
        sender = input_port[0].path_afferents[0].sender
        if not isinstance(sender.owner, CompositionInterfaceMechanism):
            return sender, sender.owner, comp
        return self._get_source_of_modulation_for_parameter_CIM(sender, sender.owner.composition)

    def _get_source_info_from_output_CIM(self, port, comp=None)->tuple:
        """Return Port, Node and Composition for "original" source of projection from **port**.
        **port** InputPort or OutputPort of the output_CIM from which the projection of interest projects;
                 used to find source (key) in output_CIM's port_map.
        **comp** Composition at which to begin the search (or continue it when called recursively);
                 assumes the current CompositionInterfaceMechanism's Composition by default.
        """
        # Ensure method is being called on an output_CIM
        assert self == self.composition.output_CIM, f"_get_source_info_from_output_CIM called on {self.name} " \
                                                    f"which is not an output_CIM"
        #  CIM MAP ENTRIES:  [SENDER PORT,  [output_CIM InputPort,  output_CIM OutputPort]]
        # Get sender to input_port of output_CIM
        comp = comp or self.composition
        port_map = port.owner.port_map
        idx = 0 if isinstance(port, InputPort) else 1
        input_ports = [port_map[k][0] for k in port_map if port_map[k][idx] is port]
        assert len(input_ports)==1, f"PROGRAM ERROR: Expected exactly 1 input_port for {port.name} " \
                                   f"in port_map for {port.owner}; found {len(input_ports)}."
        assert len(input_ports[0].path_afferents)==1, f"PROGRAM ERROR: Port ({input_ports[0].name}) expected to have " \
                                                     f"just one path_afferent; has {len(input_ports.path_afferents)}."
        sender = input_ports[0].path_afferents[0].sender
        if not isinstance(sender.owner, CompositionInterfaceMechanism):
            return sender, sender.owner, comp
        return self._get_source_info_from_output_CIM(sender, sender.owner.composition)

    def _get_destination_info_for_output_CIM(self, port, comp=None)-> list or None:
        """Return Port, Node and Composition for "ultimate" destination(s) of projection to **port**.
        **port**: InputPort or OutputPort of the output_CIM to which the projection of interest projects;
                  used to find source (key=SENDER PORT) of the projection to the output_CIM.
        **comp**: Composition at which to begin the search (or continue it when called recursively);
                 assumes the Composition for the output_CIM to which **port** belongs by default
        Return list of tuples, one for each destination, if there is more than one destination;
          this occurs if the source of the projection to the output_CIM (SENDER PORT) is a Node in a nested Composition
          that is specified to project to more than one Node in the outer Composition
        Return None if there are no destination Nodes (i.e., source is output of outermost Composition
        """
        from psyneulink.core.compositions.composition import get_composition_for_node

        # Ensure method is being called on an output_CIM
        assert self == self.composition.output_CIM
        #  CIM MAP ENTRIES:  [SENDER PORT,  [output_CIM InputPort,  output_CIM OutputPort]]
        # Get receiver of output_port of output_CIM
        comp = comp or self.composition
        port_map = port.owner.port_map
        idx = 0 if isinstance(port, InputPort) else 1
        output_ports = [port_map[k][1] for k in port_map if port_map[k][idx] is port]
        assert len(output_ports)==1, f"PROGRAM ERROR: Expected exactly 1 output_port for {port.name} " \
                                   f"in port_map for {port.owner}; found {len(output_ports)}."
        output_port = output_ports[0]
        receivers_info = []
        if not output_port.efferents:
            return None
        for efferent in output_port.efferents:
            receiver = efferent.receiver
            if not isinstance(efferent.receiver.owner, CompositionInterfaceMechanism):
                assert comp.is_nested
                receiver_comp = get_composition_for_node(receiver.owner)
                receivers_info.append((efferent.receiver, efferent.receiver.owner, receiver_comp))
            else:
                receivers_info.append(self._get_destination_info_for_output_CIM(efferent.receiver,
                                                                                efferent.receiver.owner.composition))
        return receivers_info if any(receivers_info) else None

    def _sender_is_probe(self, output_port):
        """Return True if source of output_port is a PROBE Node of the Composition to which it belongs"""
        from psyneulink.core.compositions.composition import NodeRole
        port, node, comp = self._get_source_info_from_output_CIM(output_port, self.composition)
        return NodeRole.PROBE in comp.get_roles_by_node(node)



from psyneulink import *
import numpy as np

# CONSTRUCT THE MODEL ***********************************

# Construct the color naming pathway:
color_input = ProcessingMechanism(name='COLOR INPUT', input_shapes=2) # Note:  default function is Linear
color_input_to_hidden_wts = np.array([[2, -2], [-2, 2]])
color_hidden = ProcessingMechanism(name='COLOR HIDDEN', input_shapes=2, function=Logistic(bias=-4))
color_hidden_to_output_wts = np.array([[2, -2], [-2, 2]])
output = ProcessingMechanism(name='OUTPUT', input_shapes=2, function=Logistic)
color_pathway = [color_input, color_input_to_hidden_wts, color_hidden, color_hidden_to_output_wts, output]

# Construct the word reading pathway (using the same output_layer)
word_input = ProcessingMechanism(name='WORD INPUT', input_shapes=2)
word_input_to_hidden_wts = np.array([[3, -3], [-3, 3]])
word_hidden = ProcessingMechanism(name='WORD HIDDEN', input_shapes=2, function=Logistic(bias=-4))
word_hidden_to_output_wts = np.array([[3, -3], [-3, 3]])
word_pathway = [word_input, word_input_to_hidden_wts, word_hidden, word_hidden_to_output_wts, output]

# Construct the task specification pathways
task_input = ProcessingMechanism(name='TASK INPUT', input_shapes=2)
task = LCAMechanism(name='TASK', input_shapes=2, initial_value=[0.5, 0.5])
task_color_wts = np.array([[4,4],[0,0]])
task_word_wts = np.array([[0,0],[4,4]])
task_color_pathway = [task_input, task, task_color_wts, color_hidden]
task_word_pathway = [task_input, task, task_word_wts, word_hidden]

# Construct the decision pathway:
decision = DDM(name='DECISION',
               input_format=ARRAY,
               function=DriftDiffusionAnalytical(drift_rate=(1.0),
                                                 threshold=(0.2645),
                                                 noise=(0.5),
                                                 starting_value=(0),
                                                 non_decision_time=0.15),
               output_ports=[DECISION_VARIABLE,
                              RESPONSE_TIME,
                              PROBABILITY_UPPER_THRESHOLD]
               )
decision_pathway = [output, decision]

# Reward Mechanism - should receive input from environment based on DECISION_VARIABLE of decision Mechanism
reward = TransferMechanism(name='reward')

# Construct the Composition:
# Stroop_model = Composition(name='Stroop Model', controller=control)
Stroop_model = Composition(name='Stroop Model - EVC')
Stroop_model.add_linear_processing_pathway(color_pathway)
Stroop_model.add_linear_processing_pathway(word_pathway)
Stroop_model.add_linear_processing_pathway(task_color_pathway)
Stroop_model.add_linear_processing_pathway(task_word_pathway)
Stroop_model.add_linear_processing_pathway(decision_pathway)
Stroop_model.add_node(reward)

# Assign Scheduler Conditions:
settling_time = 1
# scheduler = Scheduler(composition=Stroop_model)
Stroop_model.scheduler.add_condition(color_hidden, EveryNCalls(task, settling_time))
Stroop_model.scheduler.add_condition(word_hidden, EveryNCalls(task, settling_time))

# Construct and add ControlMechanism
control_signal_search_range = SampleSpec(start=1.0, stop=1.8, step=0.2)
evc = OptimizationControlMechanism(name='EVC',
                                   agent_rep=Stroop_model,
                                   state_features=[color_input.input_port, word_input.input_port, reward.input_port],
                                   state_feature_function=AdaptiveIntegrator(rate=1.0),
                                   # state_feature_function=AdaptiveIntegrator,
                                   objective_mechanism= \
                                       ObjectiveMechanism(
                                               name='EVC Objective Mechanism',
                                               function=LinearCombination(operation=PRODUCT),
                                               monitor=[reward,
                                                        (decision.output_ports[PROBABILITY_UPPER_THRESHOLD], 1, -1)]),
                                   function=GridSearch,
                                   control_signals=[ControlSignal(modulates=[(GAIN, color_hidden)],
                                                                  function=Linear,
                                                                  variable=1.0,
                                                                  intensity_cost_function=Exponential(rate=0.8046),
                                                                  allocation_samples=control_signal_search_range),
                                                    ControlSignal(modulates=[(GAIN, word_hidden)],
                                                                  function=Linear,
                                                                  variable=1.0,
                                                                  intensity_cost_function=Exponential(rate=0.8046),
                                                                  allocation_samples=control_signal_search_range)])
Stroop_model.add_controller(evc)

# SHOW_GRAPH ***********************************

# Stroop_model.show_graph()
Stroop_model.show_graph(show_controller=True)
# Stroop_model.show_graph(show_node_structure=ALL)
# Stroop_model.show_graph(show_dimensions=True, show_node_structure=ALL)

# RUN THE MODEL *********************************

# Label inputs
red = [1,0]
green = [0,1]
word = [0,1]
color = [1,0]

np.set_printoptions(precision=2)
global t
t = 0
def print_after():
    global t
    if t==0:
        return
    print(f'\nEnd of trial {t}:')
    print(f'\t\t\t\tcolor  word')
    print(f'\ttask:\t\t{task.value[0]}')
    print(f'\ttask gain:\t   {task.parameter_ports[GAIN].value}')
    print(f'\t\t\t\tred   green')
    print(f'\toutput:\t\t{output.value[0]}')
    print(f'\tdecision:\t{decision.value[0]}{decision.value[1]}')
    print(f'\tconflict:\t  {control._objective_mechanism.value[0]}')
    t += 1

task.log.set_log_conditions(VALUE)
evc.log.set_log_conditions(VARIABLE)
evc.log.set_log_conditions(VALUE)

task.initial_value = [0.5,0.5]
task.reset_stateful_function_when=AtPass(n=0)
# task.reset_stateful_function_when=AtTrialStart()

num_trials = 2
stimuli = {color_input:[red] * num_trials,
           word_input:[green] * num_trials,
           task_input:[color] * num_trials}
Stroop_model.run(inputs=stimuli,
                 # animate=True,
                 animate={'show_controller':True,
                          'show_cim':True},
                 call_after_trial=print_after
                 )
Stroop_model.log.print_entries()
